"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const request_promise_1 = __importDefault(require("request-promise"));
const https_1 = require("https");
const helpers_1 = require("../helpers");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const CFError = require('cf-errors');
const logger = helpers_1.createNewLogger('codefresh:infra:git-providers:bitbucket-server');
const LIMIT_PER_PAGE = 500;
const MAX_PAGES = 20;
const ApiVersions = {
    // eslint-disable-next-line @typescript-eslint/naming-convention
    V1: 'rest/api/1.0/',
};
const COMMIT_METADATA_PATH = 'com.atlassian.bitbucket.server.bitbucket-branch:latest-commit-metadata';
const statesMap = {
    pending: 'INPROGRESS',
    running: 'INPROGRESS',
    success: 'SUCCESSFUL',
    failure: 'FAILED',
    error: 'FAILED',
};
const ownerRegex = /(scm\/)?(?<noun>~)?(?<projOrUser>[^/]*)\/(?<repo>[^/]*)$/;
const _cleanEncodedFilePath = (filepath) => {
    const prefix = './';
    const hasPrefix = filepath.indexOf(prefix) === 0;
    const path = hasPrefix ? filepath.slice(prefix.length) : filepath;
    return encodeURIComponent(path);
};
const _extractErrorFromResponse = (response) => {
    let errStr = `${lodash_1.get(response, 'body.errors[0].message')}`;
    if (lodash_1.get(response, 'body.errors[0].exceptionName')) {
        errStr = `${errStr} [${lodash_1.get(response, 'body.errors[0].exceptionName')}]`;
    }
    return errStr;
};
const _toBranch = (rawBranch) => ({
    name: rawBranch.displayId,
    id: rawBranch.id,
    commit: {
        sha: rawBranch.metadata[COMMIT_METADATA_PATH].id,
        commiter_name: rawBranch.metadata[COMMIT_METADATA_PATH].author.name,
        message: rawBranch.metadata[COMMIT_METADATA_PATH].message,
        url: '',
    },
});
const _toRepo = (rawRepo, lastCommit, defaultBranch) => {
    const isPersonal = lodash_1.get(rawRepo, 'project.type') === 'PERSONAL';
    let fullName = '';
    if (isPersonal) {
        fullName = `${lodash_1.get(rawRepo, 'project.owner.name')}/${rawRepo.slug}`;
    }
    else {
        fullName = `${lodash_1.get(rawRepo, 'project.name')}/${rawRepo.slug}`;
    }
    const httpClone = lodash_1.find(lodash_1.get(rawRepo, 'links.clone', []), (l) => l.name === 'http' || l.name === 'https');
    const sshClone = lodash_1.find(lodash_1.get(rawRepo, 'links.clone', []), (l) => l.name === 'ssh');
    return {
        id: rawRepo.hierarchyId,
        provider: 'bitbucket-server',
        name: rawRepo.name,
        full_name: fullName,
        private: !rawRepo.public,
        pushed_at: (lastCommit === null || lastCommit === void 0 ? void 0 : lastCommit.committerTimestamp) && new Date(lastCommit.committerTimestamp),
        open_issues: 0,
        clone_url: lodash_1.get(httpClone, 'href', ''),
        ssh_url: lodash_1.get(sshClone, 'href', ''),
        owner: {
            login: lodash_1.get(rawRepo, isPersonal ? 'project.owner.name' : 'project.name', ''),
            avatar_url: '',
            creator: null,
        },
        org: isPersonal ? null : lodash_1.get(rawRepo, 'project.name', null),
        default_branch: lodash_1.get(defaultBranch, 'displayId', ''),
        permissions: {
            admin: true,
        },
        webUrl: lodash_1.get(rawRepo, 'links.self[0].href', ''),
    };
};
const _toWebhook = (owner, repo, rawWebhook) => {
    return {
        id: rawWebhook.id,
        name: String(rawWebhook.url),
        endpoint: rawWebhook.url,
        repo: `${owner}/${repo}`,
        events: rawWebhook.events,
    };
};
const _toUser = (user) => {
    return {
        login: user.name,
        email: user.emailAddress,
        avatar_url: user.avatarUrl,
        web_url: user.links.self[0],
    };
};
const getRepoInfo = (fullRepoName) => {
    var _a;
    const groups = (_a = fullRepoName.match(ownerRegex)) === null || _a === void 0 ? void 0 : _a.groups;
    if (!ownerRegex.test(fullRepoName) || !groups) {
        throw new CFError(`${fullRepoName} is invalid`);
    }
    return {
        noun: groups.noun || '',
        owner: groups.projOrUser,
        repo: groups.repo
    };
};
class BitbucketServer {
    constructor(opts) {
        const apiUrl = opts.apiURL || opts.apiUrl;
        if (!apiUrl) {
            throw new CFError('Cannot create bitbucket-server provider without api url');
        }
        this.baseUrl = apiUrl;
        if (!this.baseUrl.endsWith('/')) {
            this.baseUrl = `${this.baseUrl}/`;
        }
        this.timeout = opts.timeout || 10000;
        if (opts.type === 'basic') {
            this.authenticationHeader = {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                Authorization: `Basic ${Buffer.from(`${opts.username}:${opts.password}`).toString('base64')}`,
            };
        }
        else {
            // eslint-disable-next-line @typescript-eslint/naming-convention
            this.authenticationHeader = { Authorization: `Bearer ${opts.password}` };
        }
        if (apiUrl.startsWith('https') && (process.env.NODE_TLS_REJECT_UNAUTHORIZED === '0' || opts.insecure)) {
            logger.warn('using insecure mode');
            this.agent = new https_1.Agent({ rejectUnauthorized: false });
        }
    }
    async createRepository(opt) {
        const isOwner = opt.owner.startsWith('~');
        const projectKey = isOwner
            ? opt.owner
            : await this.getProjectKey(opt.owner);
        const [err, res] = await helpers_1.to(this.performAPICall({
            api: `${isOwner ? 'users' : 'projects'}/${projectKey}/repos`,
            method: 'POST',
            json: true,
            data: {
                name: opt.repo,
                scmId: 'git'
            }
        }));
        if (err) {
            throw new CFError(`Failed to create repository ${opt.repo} in ${projectKey} , err: ${JSON.stringify(err)}`);
        }
        return _toRepo(res.body);
    }
    async performAPICall(opts) {
        const method = opts.method || 'GET';
        const requestHeaders = lodash_1.merge(opts.noAuth ? {} : this.authenticationHeader, opts.headers || {});
        const requestOptions = {
            method,
            url: `${this.baseUrl}${opts.apiVersion || ApiVersions.V1}${opts.api}`,
            qs: opts.qs || {},
            headers: requestHeaders,
            json: opts.json,
            timeout: this.timeout,
            body: opts.data,
            resolveWithFullResponse: true,
            agent: this.agent,
            simple: false,
        };
        logger.debug(`${method} ${requestOptions.url} qs: ${JSON.stringify(requestOptions.qs)}`);
        return request_promise_1.default(requestOptions).then((res) => {
            const curLogger = res.statusCode >= 400
                ? logger.error.bind(logger)
                : logger.debug.bind(logger);
            curLogger(`${method} ${requestOptions.url} qs: ${JSON.stringify(requestOptions.qs)} status: ${res.statusCode}`);
            return res;
        });
    }
    async paginateForResult(opts, predicate) {
        let start = 0;
        let page = 0;
        let isLastPage = false;
        const allRes = [];
        while (!isLastPage && page < MAX_PAGES) {
            opts.qs = lodash_1.merge(opts.qs, {
                start: '' + start,
                limit: '' + LIMIT_PER_PAGE,
            });
            const res = await this.performAPICall(opts);
            if (res.statusCode >= 400) {
                throw new CFError(`Failed with status code: ${res.statusCode}, ${_extractErrorFromResponse(res)}`);
            }
            isLastPage = res.body.isLastPage;
            const predicateResult = predicate && lodash_1.find(res.body.values, predicate);
            if (predicateResult) {
                return predicateResult; // found the wanted result
            }
            allRes.push(...res.body.values);
            start += LIMIT_PER_PAGE;
            page += 1;
        }
        return predicate ? undefined : allRes;
    }
    async getProjectKey(projectName) {
        const [projectByKeyError] = await helpers_1.to(this.getProjectByKey(projectName));
        if (!projectByKeyError) {
            return projectName;
        }
        const [getProjErr, proj] = await helpers_1.to(this.paginateForResult({
            api: `projects`,
            json: true,
            qs: {
                name: projectName,
            },
        }, (project) => project.name === projectName));
        if (getProjErr) {
            throw new CFError(`Failed to get project "${projectName}", with: ${getProjErr}`);
        }
        if (!proj) {
            throw new CFError(`Project with name "${projectName}" was not found. To use the user-centric api add the "~" prefix: "~${projectName}"`);
        }
        return proj.key;
    }
    async getProjectByKey(projectKey) {
        const [getProjErr, proj] = await helpers_1.to(this.performAPICall({
            api: `projects/${projectKey}`,
            json: true,
        }));
        if (getProjErr) {
            throw new CFError(`Failed to get project with key "${projectKey}", with: ${getProjErr}`);
        }
        return proj.body;
    }
    getName() {
        return 'bitbucket-server';
    }
    async fetchRawFile(opts) {
        const projectKey = opts.owner.startsWith('~')
            ? opts.owner
            : await this.getProjectKey(opts.owner);
        const lines = [];
        let start = 0;
        let page = 0;
        let isLastPage = false;
        while (!isLastPage && page < MAX_PAGES) {
            const res = await this.performAPICall({
                apiVersion: ApiVersions.V1,
                api: `projects/${projectKey}/repos/${opts.repo}/browse/${_cleanEncodedFilePath(opts.path)}`,
                json: true,
                qs: {
                    at: opts.ref,
                    raw: 'true',
                    start: '' + start,
                    limit: '' + LIMIT_PER_PAGE,
                },
            });
            if (res.statusCode >= 400) {
                throw new CFError(`Failed to retrieve file: ${JSON.stringify(opts)}, status code: ${res.statusCode}, err: ${_extractErrorFromResponse(res)}`);
            }
            isLastPage = res.body.isLastPage;
            lines.push(...lodash_1.map(res.body.lines, (line) => line.text));
            start += LIMIT_PER_PAGE;
            page += 1;
        }
        return lines.join('\n');
    }
    async getBranch(opts) {
        const projectKey = opts.owner.startsWith('~')
            ? opts.owner
            : await this.getProjectKey(opts.owner);
        const [getBranchErr, branch] = await helpers_1.to(this.paginateForResult({
            apiVersion: ApiVersions.V1,
            api: `projects/${projectKey}/repos/${opts.repo}/branches`,
            json: true,
            qs: {
                filterText: opts.branch,
                details: 'true',
            },
        }, (b) => lodash_1.isEqual(opts.branch, b.displayId ? b.displayId : lodash_1.last(b.id.split('/')))));
        if (getBranchErr) {
            throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opts)}, ${getBranchErr}`);
        }
        if (!branch) {
            throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opts)}, branch not found`);
        }
        return _toBranch(branch);
    }
    async getRepository(opts) {
        const projectKey = opts.owner.startsWith('~')
            ? opts.owner
            : await this.getProjectKey(opts.owner);
        const [repo, lastCommit, defaultBranch] = await Promise.all([
            this.performAPICall({
                apiVersion: ApiVersions.V1,
                api: `projects/${projectKey}/repos/${opts.repo}`,
                json: true,
            }),
            this.performAPICall({
                apiVersion: ApiVersions.V1,
                api: `projects/${projectKey}/repos/${opts.repo}/commits`,
                json: true,
            }),
            this.performAPICall({
                apiVersion: ApiVersions.V1,
                api: `projects/${projectKey}/repos/${opts.repo}/branches/default`,
                json: true,
            }),
        ]);
        if (repo.statusCode >= 400) {
            throw new CFError(`Failed to get repository: ${JSON.stringify(opts)}, status code: ${repo.statusCode}, err: ${_extractErrorFromResponse(repo)}`);
        }
        if (lastCommit.statusCode >= 400) {
            throw new CFError(`Failed to get repository lastest commit: ${JSON.stringify(opts)}, status code: ${lastCommit.statusCode}, err: ${_extractErrorFromResponse(lastCommit)}`);
        }
        if (defaultBranch.statusCode >= 400) {
            throw new CFError(`Failed to get repository default branch: ${JSON.stringify(opts)}, status code: ${defaultBranch.statusCode}, err: ${_extractErrorFromResponse(defaultBranch)}`);
        }
        return _toRepo(repo.body, lodash_1.get(lastCommit.body, 'values[0]', {}), defaultBranch.body);
    }
    async listBranches(opt) {
        throw new Error('Method not implemented.');
    }
    async listRepositoriesForOwner(opt) {
        throw new Error('Method not implemented.');
    }
    async createRepositoryWebhook(opt) {
        const projectKey = opt.owner.startsWith('~')
            ? opt.owner
            : await this.getProjectKey(opt.owner);
        const [err, res] = await helpers_1.to(this.performAPICall({
            api: `projects/${projectKey}/repos/${opt.repo}/webhooks`,
            method: 'POST',
            json: true,
            data: {
                events: [
                    'repo:refs_changed',
                    'pr:opened',
                    'pr:merged',
                    'pr:modified',
                    'pr:deleted',
                    'pr:declined',
                    'pr:reviewer:unapproved',
                    'pr:reviewer:approved',
                    'pr:reviewer:needs_work',
                    'pr:comment:edited',
                    'pr:comment:added',
                    'pr:comment:deleted',
                ],
                configuration: {
                    secret: opt.secret,
                },
                url: opt.endpoint,
                active: true,
            },
        }));
        if (err) {
            throw new CFError(`Failed to create repository webhooks: ${JSON.stringify(opt)}, err: ${JSON.stringify(err)}`);
        }
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to create repository webhooks: ${JSON.stringify(opt)}, status code: ${res.statusCode},  err: ${_extractErrorFromResponse(res)}`);
        }
        return _toWebhook(opt.owner, opt.repo, res.body);
    }
    async listWebhooks(opt) {
        const projectKey = opt.owner.startsWith('~')
            ? opt.owner
            : await this.getProjectKey(opt.owner);
        const [getHooksErr, hooks] = await helpers_1.to(this.paginateForResult({
            api: `projects/${projectKey}/repos/${opt.repo}/webhooks`,
            json: true,
        }));
        if (getHooksErr) {
            throw new CFError(`Failed to list repository webhooks: ${JSON.stringify(opt)}, err: ${JSON.stringify(getHooksErr)}`);
        }
        if (hooks.statusCode >= 400) {
            throw new CFError(`Failed to list repository webhooks: ${JSON.stringify(opt)}, status code: ${hooks.statusCode},  err: ${_extractErrorFromResponse(hooks)}`);
        }
        return hooks.map(_toWebhook.bind(null, opt.owner, opt.repo));
    }
    async deleteRepositoryWebhook(opt) {
        const projectKey = opt.owner.startsWith('~')
            ? opt.owner
            : await this.getProjectKey(opt.owner);
        const [err, res] = await helpers_1.to(this.performAPICall({
            api: `projects/${projectKey}/repos/${opt.repo}/webhooks/${opt.hookId}`,
            method: 'DELETE',
            json: true,
        }));
        if (err) {
            throw new CFError(`Failed to delete webhook: ${JSON.stringify(opt)}, err: ${JSON.stringify(err)}`);
        }
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to delete webhook: ${JSON.stringify(opt)}, status code: ${res.statusCode}, err: ${_extractErrorFromResponse(res)}`);
        }
    }
    async listRepositoriesWithAffiliation(opt) {
        const [getReposErr, repos] = await helpers_1.to(this.paginateForResult({
            api: `repos`,
            json: true,
            qs: {
                permission: 'REPO_READ'
            }
        }));
        if (getReposErr) {
            throw new CFError(`Failed to list repos: ${JSON.stringify(opt)}, err: ${JSON.stringify(getReposErr)}`);
        }
        if (repos.statusCode >= 400) {
            throw new CFError(`Failed to get repos: ${JSON.stringify(opt)}, status code: ${repos.statusCode}, ${_extractErrorFromResponse(repos)}`);
        }
        return repos.map(_toRepo);
    }
    async listOrganizations(opt) {
        const [err, res] = await helpers_1.to(this.performAPICall({
            api: `projects`,
            json: true,
        }));
        if (err) {
            throw new CFError({
                message: `Failed to list organization caused by ${err}`,
                cause: err,
            });
        }
        // no organizations
        if (!res.body.values || res.body.values.length == 0) {
            return [];
        }
        return res.body.values.map((project) => project.name);
    }
    async getRepositoryPermissions(opt) {
        var _a, _b, _c, _d;
        const projectKey = opt.owner.startsWith('~')
            ? opt.owner
            : await this.getProjectKey(opt.owner);
        const [reposReadResErr, reposReadRes] = await helpers_1.to(this.performAPICall({
            apiVersion: ApiVersions.V1,
            api: `projects/${projectKey}/repos?permissions=REPO_READ`,
            json: true,
        }));
        if (reposReadResErr) {
            throw new CFError({
                message: `Failed to get repository permissions: ${opt.owner}/${opt.repo}: ${reposReadResErr}`,
                cause: reposReadResErr,
            });
        }
        const read = -1 != ((_b = (_a = reposReadRes.body) === null || _a === void 0 ? void 0 : _a.values) === null || _b === void 0 ? void 0 : _b.findIndex((repo) => repo.name == opt.repo));
        const [reposWriteResErr, reposWriteRes] = await helpers_1.to(this.performAPICall({
            apiVersion: ApiVersions.V1,
            api: `projects/${projectKey}/repos?permissions=REPO_WRITE`,
            json: true,
        }));
        if (reposWriteResErr) {
            throw new CFError({
                message: `Failed to get repository permissions: ${opt.owner}/${opt.repo}: ${reposWriteResErr}`,
                cause: reposWriteResErr,
            });
        }
        const write = -1 != ((_d = (_c = reposWriteRes.body) === null || _c === void 0 ? void 0 : _c.values) === null || _d === void 0 ? void 0 : _d.findIndex((repo) => repo.name == opt.repo));
        return {
            read,
            write
        };
    }
    async listRepositoriesForOrganization(opt) {
        throw new Error('Method not implemented.');
    }
    async createCommitStatus(opt) {
        const projectKey = opt.owner.startsWith('~')
            ? opt.owner
            : await this.getProjectKey(opt.owner);
        const [err, res] = await helpers_1.to(this.performAPICall({
            api: `projects/${projectKey}/repos/${opt.repo}/commits/${opt.sha}/builds`,
            method: 'POST',
            json: true,
            data: {
                key: opt.context,
                name: opt.context,
                description: opt.description,
                url: opt.targetUrl,
                state: statesMap[opt.state],
            },
        }));
        if (err) {
            throw new CFError(`Failed to create commit status: ${JSON.stringify(opt)}, err: ${JSON.stringify(err)}`);
        }
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to create commit status: ${JSON.stringify(opt)}, status code: ${res.statusCode}, err: ${_extractErrorFromResponse(res)}`);
        }
    }
    async getUser(opt) {
        var _a;
        let username = opt === null || opt === void 0 ? void 0 : opt.username;
        if (!username) {
            const [err, res] = await helpers_1.to(this.performAPICall({
                api: `whoami`,
                method: 'GET',
                json: true,
                // ApiVersions.V1 dosent supply a whoami endpoint
                apiVersion: `plugins/servlet/applinks/`
            }));
            if (err) {
                throw new CFError(`Failed to get current user, err: ${JSON.stringify(err)}`);
            }
            username = res.body;
        }
        const [err, res] = await helpers_1.to(this.performAPICall({
            // avatarSize adds avatar URLs for commit authors and specifying the desired size in pixels
            api: `users/${username}?avatarSize=64`,
            method: 'GET',
            json: true,
        }));
        if (err) {
            throw new CFError(`Failed to get ${username}, err: ${JSON.stringify(err)}`);
        }
        if ((_a = res.body) === null || _a === void 0 ? void 0 : _a.errors) {
            throw new CFError(`Failed to get ${username}, err: ${JSON.stringify(res.body.errors)}`);
        }
        return _toUser(res.body);
    }
    async getUserByEmail(email) {
        const [err, res] = await helpers_1.to(this.performAPICall({
            // avatarSize adds avatar URLs for commit authors and specifying the desired size in pixels
            api: `users?avatarSize=64`,
            method: 'GET',
            json: true,
        }));
        if (err) {
            throw new CFError(`Failed to get user connected to email ${email} , err: ${JSON.stringify(err)}`);
        }
        const user = res.body.values.find((user) => user.emailAddress == email);
        if (!user) {
            throw new CFError(`User with email ${email} was not found`);
        }
        return _toUser(user);
    }
    async getPullRequestFiles(opt) {
        throw new Error('Method not implemented.');
    }
    async getPullRequest(opt) {
        throw new Error('Method not implemented.');
    }
    async assertApiScopes(opt) {
        // bitbucket server doesnt supply endpoint with permissions scopes info
        // there is no enpoint for write only that is not project admin
        var _a;
        const [getProjErr, projects] = await helpers_1.to(this.performAPICall({
            api: `projects`,
            method: 'GET',
            json: true,
        }));
        if (getProjErr) {
            throw new CFError({
                message: `ValidationError: failed to get project, caused by err: ${JSON.stringify(getProjErr)}`,
                cause: getProjErr
            });
        }
        if (!opt.scopes.includes('repo_write')) {
            return;
        }
        if (!projects.body.values || projects.body.values.length === 0) {
            throw new CFError({
                message: `ValidationError: no project to check permissions for`,
                cause: getProjErr
            });
        }
        // this request allowed only for PROJECT_ADMIN only
        // which is repo_read and repo_write global scopes
        const projectKey = (_a = projects.body.values[0]) === null || _a === void 0 ? void 0 : _a.key;
        const [getPermissionsErr] = await helpers_1.to(this.performAPICall({
            api: `projects/${projectKey}/permissions/groups`,
            method: 'GET',
            json: true,
        }));
        if (getPermissionsErr) {
            throw new CFError({
                message: `ValidationError: check your token permissions, expected: ${opt.scopes.toString()}`,
                cause: getPermissionsErr
            });
        }
    }
    toOwnerRepo(fullRepoName) {
        const repoInfo = getRepoInfo(fullRepoName);
        const repo = repoInfo.repo;
        const owner = repoInfo.owner;
        const noun = repoInfo.noun;
        if (!owner || !repo) {
            throw new CFError(`Failed to get repo and owner from repository name ${fullRepoName}`);
        }
        return [noun + owner, repo];
    }
    getAuth() {
        return {
            headers: this.authenticationHeader,
        };
    }
}
exports.default = BitbucketServer;
//# sourceMappingURL=bitbucket-server.js.map